#
# Copyright (C) 2012 Marlus Saraiva, Rodrigo Maia
# Copyright (C) 2019 Stefan Eilers
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

require "jasper-rails/version"
require "rails"
require "rjb"
#require "action_controller/metal/responder"

if Mime::Type.lookup_by_extension("pdf").nil?
    Mime::Type.register "application/pdf", :pdf
end

module JasperRails

    class << self
        attr_accessor :config
    end

    module Jasper
        module Rails

            def self.classpath
                classpaths = '.'
                Dir["#{File.dirname(__FILE__)}/java/*.jar"].each do |jar|
                    classpaths << File::PATH_SEPARATOR + File.expand_path(jar)
                end

                Dir["lib/*.jar"].each do |jar|
                    classpaths << File::PATH_SEPARATOR + File.expand_path(jar)
                end
                classpaths
            end

            ##
            ## @brief      Render a PDF file based on jasper file and an optional datasource.
            ##
            ## @param      jasper_file  The Jasper-Report Template (.jrxml)
            ## @param      datasource   The Datasource (JRAbstractTextDataSource)
            ## @param      parameters   The parameters (Ruby-Hash)
            ##
            ## @return     The PDF-Data
            ##
            def self.render_pdf(jasper_file, parameters, datasource = nil, create_pdf_a = false )
                setup_jasper
                # The code below is to workaround declaring constants within methods
                # We would like to delay these till a request is received to workaround Apache Passenger issue
                cLocale                      		= Rjb::import 'java.util.Locale'
                cJasperCompileManager        		= Rjb::import 'net.sf.jasperreports.engine.JasperCompileManager'
                cJasperExportManager         		= Rjb::import 'net.sf.jasperreports.engine.JasperExportManager'
                cJasperFillManager           		= Rjb::import 'net.sf.jasperreports.engine.JasperFillManager'
                cJREmptyDataSource           		= Rjb::import 'net.sf.jasperreports.engine.JREmptyDataSource'
                cHashMap                     		= Rjb::import 'java.util.HashMap'
                cJavaString                  		= Rjb::import 'java.lang.String'

                # PDF/A Export
                cJRPdfExporter 						= Rjb::import 'net.sf.jasperreports.engine.export.JRPdfExporter'
                cSimpleExporterInput         		= Rjb::import 'net.sf.jasperreports.export.SimpleExporterInput'
                cSimpleOutputStreamExporterOutput   = Rjb::import 'net.sf.jasperreports.export.SimpleOutputStreamExporterOutput'
                cSimplePdfExporterConfiguration 	= Rjb::import 'net.sf.jasperreports.export.SimplePdfExporterConfiguration'
                cPdfaConformanceEnum 				= Rjb::import 'net.sf.jasperreports.export.type.PdfaConformanceEnum'
                cByteArrayOutputStream				= Rjb::import 'java.io.ByteArrayOutputStream'
                cDefaultJasperReportsContext		= Rjb::import 'net.sf.jasperreports.engine.DefaultJasperReportsContext'

                options ||= {}
                parameters ||= {}
                jrxml_file  = jasper_file.sub(/\.jasper$/, ".jrxml")

                # Basic setup of JasperReports
                if create_pdf_a
                    jasperReportsContext = cDefaultJasperReportsContext.getInstance();
                    # Overwrite default fonts in order to get an embedded report
                    # TODO: This might change the report. Try following: https://medium.com/@seymorethrottle/jasper-reports-adding-custom-fonts-589b55a52e7c
                    jasperReportsContext.setProperty("net.sf.jasperreports.default.pdf.font.name", "net/sf/jasperreports/fonts/dejavu/DejaVuSans.ttf")
                    jasperReportsContext.setProperty("net.sf.jasperreports.default.pdf.embedded", "true");
                end

                begin
                    # Default report params
                    config = {
                        :report_params=>{
                            "REPORT_LOCALE"    => cLocale.new('de', 'DE'),
                            "XML_LOCALE"       => cLocale.new('de', 'DE'),
                            "XML_DATE_PATTERN" => 'yyyy-MM-dd'
                        }
                    }

                    # Converting default report params to java HashMap
                    jasper_params = cHashMap.new
                    config[:report_params].each do |k,v|
                        jasper_params.put(k, v)
                    end

                    # Convert the ruby parameters' hash to a java HashMap, but keeps it as
                    # default when they already represent a JRB entity.
                    # Pay attention that, for now, all other parameters are converted to string!
                    parameters.each do |key, value|
                        jasper_params.put(cJavaString.new(key.to_s), parameter_value_of(value))
                    end

                    # Compile it, if needed
                    if !File.exist?(jasper_file) || (File.exist?(jrxml_file) && File.mtime(jrxml_file) > File.mtime(jasper_file))
                        cJasperCompileManager.compileReportToFile(jrxml_file, jasper_file)
                    end

                    # Fill the report
                    if datasource
                        jasper_print = cJasperFillManager.fillReport(jasper_file, jasper_params, datasource || cJREmptyDataSource.new)
                    else
                        jasper_print = cJasperFillManager.fillReport(jasper_file, jasper_params)
                    end

                    if create_pdf_a
                        ## Export as PDF/A with embedded fonts
                        icc_path = "#{File.dirname(__FILE__)}/java/sRGB_v4_ICC_preference_displayclass.icc"
                        exporterConfiguration = cSimplePdfExporterConfiguration.new;
                        exporterConfiguration.isTagged()
                        exporterConfiguration.setPdfaConformance(cPdfaConformanceEnum.PDFA_1A);
                        exporterConfiguration.setIccProfilePath(cJavaString.new(icc_path))

                        #  JRPdfExporter exporter = new JRPdfExporter();
                        byteArrayOutputStream = cByteArrayOutputStream.new
                        exporter = cJRPdfExporter.new()
                        exporter.setExporterInput(cSimpleExporterInput.new(jasper_print))
                        exporter.setConfiguration(exporterConfiguration)
                        exporter.setExporterOutput(cSimpleOutputStreamExporterOutput.new(byteArrayOutputStream));
                        exporter.exportReport();

                        byteArrayOutputStream.toByteArray()
                    else
                        ## Export as usual PDF-File
                        cJasperExportManager._invoke('exportReportToPdf', 'Lnet.sf.jasperreports.engine.JasperPrint;', jasper_print)
                    end
                rescue Exception=>e
                    if e.respond_to? 'printStackTrace'
                        ::Rails.logger.error e.message
                        e.printStackTrace
                    else
                        ::Rails.logger.error e.message + "\n " + e.backtrace.join("\n ")
                    end
                    raise e
                end
            end

            ##
            ## @brief      Render a Jasper-Report by using a JsonDataSource
            ##
            ## @param      jasper_file  The jasper file (.jrxml)
            ## @param      parameters   The parameters (Ruby-Hash)
            ## @param      json_data    The json data (JSON-String)
            ##
            ## @return     { description_of_the_return_value }
            ##
            def self.render_json_to_pdf(jasper_file, parameters, json_data, create_pdf_a = false)
                setup_jasper

                # Convert json string to byte array.
                jsonDataStream = byteArrayInputStream.new(cJavaString.new(json_data).getBytes());
                data_source = jsonDataSource.new(jsonDataStream)

                render_pdf(jasper_file, parameters, data_source, create_pdf_a)
            end

            ##
            ## @brief      Render a report that contains a datasource.
            ##
            ## @param      jasper_file  The jasper file
            ## @param      parameters   The parameters
            ##
            ## @return     The rendered pdf file.
            ##
            def self.render_datasource_to_pdf(jasper_file, parameters, create_pdf_a = false)
                setup_jasper

                render_pdf(jasper_file, parameters, nil, create_pdf_a)
            end

            ##
            ## @brief      Render multiple fileformats based on jasper file and an optional datasource.
            ##
            ## @param      jasper_file  The Jasper-Report Template (.jrxml)
            ## @param      datasource   The Datasource (JRAbstractTextDataSource)
            ## @param      parameters   The parameters (Ruby-Hash)
            ##
            ## @return     The Data in Formats (...)
            ##
            def self.render_format(outputformat,jasper_file, parameters, datasource = nil )
                setup_jasper
                # The code below is to workaround declaring constants within methods
                # We would like to delay these till a request is received to workaround Apache Passenger issue
                cLocale                             = Rjb::import 'java.util.Locale'
                cJasperCompileManager               = Rjb::import 'net.sf.jasperreports.engine.JasperCompileManager'
                cJasperExportManager                = Rjb::import 'net.sf.jasperreports.engine.JasperExportManager'
                cJasperFillManager                  = Rjb::import 'net.sf.jasperreports.engine.JasperFillManager'
                cJREmptyDataSource                  = Rjb::import 'net.sf.jasperreports.engine.JREmptyDataSource'
                cHashMap                            = Rjb::import 'java.util.HashMap'
                cJavaString                         = Rjb::import 'java.lang.String'

                cByteArrayOutputStream              = Rjb::import 'java.io.ByteArrayOutputStream'
                cSimpleExporterInput                = Rjb::import 'net.sf.jasperreports.export.SimpleExporterInput'
                cSimpleOutputStreamExporterOutput   = Rjb::import 'net.sf.jasperreports.export.SimpleOutputStreamExporterOutput'
                cSimpleWriterExporterOutput         = Rjb::import 'net.sf.jasperreports.export.SimpleWriterExporterOutput'

                # xlsx
                cJRXlsxExporter                     = Rjb::import 'net.sf.jasperreports.engine.export.ooxml.JRXlsxExporter'
                cSimpleXlsxReportConfiguration      = Rjb::import 'net.sf.jasperreports.export.SimpleXlsxReportConfiguration'

                # docx
                cJRDocxExporter                     = Rjb::import 'net.sf.jasperreports.engine.export.ooxml.JRDocxExporter'
                cSimpleDocxReportConfiguration      = Rjb::import 'net.sf.jasperreports.export.SimpleDocxReportConfiguration'

                #html
                cHtmlExporter                       = Rjb::import 'net.sf.jasperreports.engine.export.HtmlExporter'
                cSimpleHtmlExporterOutput           = Rjb::import 'net.sf.jasperreports.export.SimpleHtmlExporterOutput'

                #csv
                cJRCsvExporter                      = Rjb::import 'net.sf.jasperreports.engine.export.JRCsvExporter'

                #pdf
                cJRPdfExporter                      = Rjb::import 'net.sf.jasperreports.engine.export.JRPdfExporter'

                #pdfA
                cSimplePdfExporterConfiguration     = Rjb::import 'net.sf.jasperreports.export.SimplePdfExporterConfiguration'
                cPdfaConformanceEnum                = Rjb::import 'net.sf.jasperreports.export.type.PdfaConformanceEnum'


                #odt
                cJROdtExporter                      = Rjb::import 'net.sf.jasperreports.engine.export.oasis.JROdtExporter'

                #ods
                cJROdsExporter                      = Rjb::import 'net.sf.jasperreports.engine.export.oasis.JROdsExporter'

                #pptx
                cJRPptxExporter                     = Rjb::import 'net.sf.jasperreports.engine.export.ooxml.JRPptxExporter'


                jBoolean = Rjb::import("java.lang.Boolean")

                # Default report params
                config = {
                    :report_params=>{
                        "REPORT_LOCALE"    => cLocale.new('de', 'DE'),
                        "XML_LOCALE"       => cLocale.new('de', 'DE'),
                        "XML_DATE_PATTERN" => 'yyyy-MM-dd'
                    }
                }

                # Converting default report params to java HashMap
                jasper_params = cHashMap.new
                config[:report_params].each do |k,v|
                    jasper_params.put(k, v)
                end

                options ||= {}
                parameters ||= {}
                jrxml_file  = jasper_file.sub(/\.jasper$/, ".jrxml")

                # Convert the ruby parameters' hash to a java HashMap, but keeps it as
                # default when they already represent a JRB entity.
                # Pay attention that, for now, all other parameters are converted to string!
                parameters.each do |key, value|
                    jasper_params.put(cJavaString.new(key.to_s), parameter_value_of(value))
                end

                # Compile it, if needed
                if !File.exist?(jasper_file) || (File.exist?(jrxml_file) && File.mtime(jrxml_file) > File.mtime(jasper_file))
                    cJasperCompileManager.compileReportToFile(jrxml_file, jasper_file)
                end

                # Fill the report
                if datasource
                    jasper_print = cJasperFillManager.fillReport(jasper_file, jasper_params, datasource || cJREmptyDataSource.new)
                else
                    jasper_print = cJasperFillManager.fillReport(jasper_file, jasper_params)
                end

                byteArrayOutputStream = cByteArrayOutputStream.new

                exporter = nil

                case outputformat
                when 'docx'
                    exporter = cJRDocxExporter.new();
                    exporter.setExporterInput(cSimpleExporterInput.new(jasper_print))
                    exporter.setExporterOutput(cSimpleOutputStreamExporterOutput.new(byteArrayOutputStream));
                    configuration = cSimpleDocxReportConfiguration.new();
                    #config.setFlexibleRowHeight(true); #Set desired configuration
                when 'xlsx'
                    exporter = cJRXlsxExporter.new();
                    exporter.setExporterInput(cSimpleExporterInput.new(jasper_print))
                    exporter.setExporterOutput(cSimpleOutputStreamExporterOutput.new(byteArrayOutputStream));
                    configuration = cSimpleXlsxReportConfiguration.new()

                    # Commented the detectcelltype, because in the Stundenzettel-Report it cannot detect the cells correctly.
                    # Excel throws an error then. 26.03.2019 LS
                    # configuration.setDetectCellType(jBoolean.TRUE) #Set configuration as you like it!!

                    # Commented because I could see a difference in the Stundenzettel Report. LS 26.03.2019
                    # configuration.setCollapseRowSpan(jBoolean.FALSE)
                    exporter.setConfiguration(configuration)
                when 'html'
                    exporter = cHtmlExporter.new()
                    exporter.setExporterInput(cSimpleExporterInput.new(jasper_print))
                    exporter.setExporterOutput(cSimpleHtmlExporterOutput.new(byteArrayOutputStream))
                when 'csv'
                    exporter = cJRCsvExporter.new()
                    exporter.setExporterInput(cSimpleExporterInput.new(jasper_print))
                    exporter.setExporterOutput(cSimpleWriterExporterOutput.new(byteArrayOutputStream));
                when 'pdf'
                    exporter = cJRPdfExporter.new();
                    exporter.setExporterInput(cSimpleExporterInput.new(jasper_print))
                    exporter.setExporterOutput(cSimpleOutputStreamExporterOutput.new(byteArrayOutputStream));
                when 'pdfa'
                   ## Export as PDF/A with embedded fonts
                    icc_path = "#{File.dirname(__FILE__)}/java/sRGB_v4_ICC_preference_displayclass.icc"
                    exporterConfiguration = cSimplePdfExporterConfiguration.new;
                    exporterConfiguration.isTagged()
                    exporterConfiguration.setPdfaConformance(cPdfaConformanceEnum.PDFA_1A);
                    exporterConfiguration.setIccProfilePath(cJavaString.new(icc_path))

                    exporter = cJRPdfExporter.new()
                    exporter.setExporterInput(cSimpleExporterInput.new(jasper_print))
                    exporter.setConfiguration(exporterConfiguration)
                    exporter.setExporterOutput(cSimpleOutputStreamExporterOutput.new(byteArrayOutputStream));
                when 'odt'
                    exporter = cJROdtExporter.new();
                    exporter.setExporterInput(cSimpleExporterInput.new(jasper_print))
                    exporter.setExporterOutput(cSimpleOutputStreamExporterOutput.new(byteArrayOutputStream));
                when 'ods'
                    exporter = cJROdtExporter.new();
                    exporter.setExporterInput(cSimpleExporterInput.new(jasper_print))
                    exporter.setExporterOutput(cSimpleOutputStreamExporterOutput.new(byteArrayOutputStream));
                when 'pptx'
                    exporter = cJRPptxExporter.new()
                    exporter.setExporterInput(cSimpleExporterInput.new(jasper_print))
                    exporter.setExporterOutput(cSimpleOutputStreamExporterOutput.new(byteArrayOutputStream));
                end

                exporter.exportReport()
                byteArrayOutputStream.toByteArray()
            end

            ##
            ## @brief      Render a Jasper-Report by using a JsonDataSource
            ##
            ## @param      jasper_file  The jasper file (.jrxml)
            ## @param      parameters   The parameters (Ruby-Hash)
            ## @param      json_data    The json data (JSON-String)
            ##
            ## @return     { description_of_the_return_value }
            ##
            def self.render_json_to_format(outputformat, jasper_file, parameters, json_data)
                setup_jasper

                # Convert json string to byte array.
                jsonDataStream = byteArrayInputStream.new(cJavaString.new(json_data).getBytes());
                data_source = jsonDataSource.new(jsonDataStream)

                render_format(outputformat, jasper_file, parameters, data_source)
            end

            ##
            ## @brief      Render a report that contains a datasource.
            ##
            ## @param      jasper_file  The jasper file
            ## @param      parameters   The parameters
            ##
            ## @return     The rendered docx file.
            ##
            def self.render_datasource_to_format(outputformat,jasper_file, parameters)
                setup_jasper

                render_format(outputformat, jasper_file, parameters, nil)
            end


            def self.mimetype_description_for_fileending(output_format)
                mimetype = ""
                description = ""
                case output_format
                when 'docx'
                    description = "MS Word (.docx)"
                    mimetype = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
                when 'xlsx'
                    description = "MS Excel (.xlsx)"
                    mimetype = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
                when 'pptx'
                    description = "MS PowerPoint (.pptx)"
                    mimetype = "application/vnd.openxmlformats-officedocument.presentationml.presentation"
                when 'odt'
                    description = "Open Office Text (.odt)"
                    mimetype = "application/vnd.oasis.opendocument.text"
                when 'ods'
                    description = "Open Office Spreadsheet (.odt)"
                    mimetype = "application/vnd.oasis.opendocument.spreadsheet"
                when 'html'
                    description = "HTML"
                    mimetype = "text/html"
                when 'csv'
                    description = "CSV"
                    mimetype = "text/comma-separated-values"
                when 'pdf','pdfa'
                    description = "PDF"
                    mimetype = "application/pdf"
                end
                return mimetype, description
            end

            def self.available_outputformats()
                available_outputformats = ['docx','xlsx','pptx','odt','ods','html','csv','pdf','pdfa']
                formats_with_mimetypes = []
                available_outputformats.each do |format|
                    mime, desc = mimetype_description_for_fileending(format)
                    format_mimetype = { "file_extension" => format , "mime" => mime, "description" => desc}
                    formats_with_mimetypes.push(format_mimetype)
                end
                formats_with_mimetypes
            end


            private

            def self.setup_jasper
                ENV['CLASS_PATH'] ||= classpath
                ENV['JVM_ARGS'] ||= '-Djava.awt.headless=true,-Xms128M,-Xmx256M'

                Rjb::load( ENV['CLASS_PATH'], ENV['JVM_ARGS'].split(",") ) unless Rjb::loaded?
            end

            # Returns the value without conversion when it's converted to Java Types.
            # When isn't a Rjb class, returns a Java String of it.
            def self.parameter_value_of(param)
                cJavaString                  = Rjb::import 'java.lang.String'
                # Using Rjb::import('java.util.HashMap').new, it returns an instance of
                # Rjb::Rjb_JavaProxy, so the Rjb_JavaProxy parent is the Rjb module itself.
                if param.class.parent == Rjb
                    param
                else
                    cJavaString.new(param.to_s)
                end
            end
        end
    end
end
